==> epic4pre0.030.tar.gz: ircii-EPIC4pre0.030/source/expr2.c <==
/*
 * A new math parser, such as it is.
 * This is still a work in progress.  It is provided
 * for your amusement and comments if you want.
 *
 * Written by Jeremy Nelson
 * Copyright 1997 EPIC Software Labs
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
char     *MatchingBracket(char *string, char left, char right);
#define LEFT_BRACKET '['
#define RIGHT_BRACKET ']'
#define LEFT_PAREN '('
#define RIGHT_PAREN ')'

#define my_isdigit(x) \
(*x >= '0' && *x <= '9') || \
((*x == '-'  || *x == '+') && (x[1] >= '0' && x[1] <= '9'))

typedef enum _token_type
{
	TOKEN_UNKNOWN,
	TOKEN_TEXT,
	TOKEN_VAR,
	TOKEN_NUMBER,
	TOKEN_OPERATOR,
	TOKEN_FUNCTION_CALL
} TOKEN_TYPE;

typedef enum _operator_type
{
	OP_UNKNOWN,
	OP_PLUS,	OP_MINUS,	OP_INCREMENT, 	OP_DECREMENT,
	OP_PLUSEQ,	OP_MINUSEQ,

	OP_MULT,	OP_DIV,		OP_MOD,		
	OP_MULTEQ,	OP_DIVEQ,	OP_MODEQ,

	OP_STRAPP,	OP_STRCAT,

	OP_BITAND,	OP_LOGAND,	OP_BITANDEQ,
	OP_BITOR,	OP_LOGOR,	OP_BITOREQ,
	OP_BITXOR,	OP_LOGXOR,	OP_BITXOREQ,

	OP_EQ,		OP_NEQ,		OP_ASSN,
	OP_LT,		OP_LTEQ,	OP_GT,		OP_GTEQ,
	OP_COMP,	OP_NOT,

	OP_TERT1,	OP_TERT2,	OP_COMMA,
	OP_LPAREN,	OP_RPAREN
} OP_TYPE;

#define	STAGE_NONE	0
#define STAGE_PAREN	1
#define	STAGE_COMMA	2
#define	STAGE_ASSN	3
#define	STAGE_TERT	4
#define	STAGE_CONJ	5
#define	STAGE_BITW	6
#define	STAGE_COMP	7
#define	STAGE_ADD	8
#define	STAGE_MULT	9

#define UNARY 1
#define ASSOC_BOTH 2
#define ASSOC_LEFT 4

struct __opvals
{
	char 		*op_str;
	OP_TYPE  	op_type;
	int		stage_type;
	int		flags;
} operators[] = 
{
	{ "",		OP_UNKNOWN,	STAGE_NONE,	0 },
	{ "+",		OP_PLUS,	STAGE_ADD,	0 },
	{ "-",		OP_MINUS,	STAGE_ADD,	0 },
	{ "++",		OP_INCREMENT,	STAGE_NONE,	UNARY | ASSOC_BOTH },
	{ "--",		OP_INCREMENT,	STAGE_NONE, 	UNARY | ASSOC_BOTH },
	{ "+=",		OP_PLUSEQ,	STAGE_ASSN, 	ASSOC_LEFT },
	{ "-=",		OP_MINUSEQ,	STAGE_ASSN, 	ASSOC_LEFT },
	{ "*",		OP_MULT,	STAGE_MULT, 	0 },
	{ "/",		OP_DIV,		STAGE_MULT, 	0 },
	{ "%",		OP_MOD,		STAGE_MULT, 	0 },
	{ "*=",		OP_MULTEQ,	STAGE_ASSN, 	ASSOC_LEFT },
	{ "/=",		OP_DIVEQ,	STAGE_ASSN, 	ASSOC_LEFT },
	{ "%=",		OP_MODEQ,	STAGE_ASSN, 	ASSOC_LEFT },
	{ "##",		OP_STRAPP,	STAGE_ADD, 	0 },
	{ "#=",		OP_STRCAT,	STAGE_ASSN, 	ASSOC_LEFT },
	{ "&",		OP_BITAND,	STAGE_BITW, 	0 },
	{ "&&",		OP_LOGAND,	STAGE_CONJ, 	0 },
	{ "&=",		OP_BITANDEQ,	STAGE_ASSN, 	ASSOC_LEFT },
	{ "|",		OP_BITOR,	STAGE_BITW, 	0 },
	{ "||",		OP_LOGOR,	STAGE_CONJ, 	0 },
	{ "|=",		OP_BITOREQ,	STAGE_ASSN, 	ASSOC_LEFT },
	{ "^",		OP_BITXOR,	STAGE_BITW, 	0 },
	{ "^^",		OP_LOGXOR,	STAGE_CONJ, 	0 },
	{ "^=",		OP_BITXOREQ,	STAGE_ASSN, 	ASSOC_LEFT },
	{ "==",		OP_EQ,		STAGE_COMP, 	0 },
	{ "!=",		OP_NEQ,		STAGE_COMP, 	0 },
	{ "=",		OP_ASSN,	STAGE_ASSN, 	ASSOC_LEFT },
	{ "<",		OP_LT,		STAGE_COMP, 	0 },
	{ ">",		OP_GT,		STAGE_COMP, 	0 },
	{ "<=",		OP_LTEQ,	STAGE_COMP, 	0 },
	{ ">=",		OP_GTEQ,	STAGE_COMP, 	0 },
	{ "~",		OP_COMP,	STAGE_BITW, 	UNARY },
	{ "!",		OP_COMP,	STAGE_CONJ,	UNARY },
	{ "?",		OP_TERT1,	STAGE_TERT,	0 },
	{ ":",		OP_TERT2,	STAGE_TERT,	0 },
	{ ",",		OP_COMMA,	STAGE_COMMA,	0 },
	{ "(",		OP_LPAREN,	STAGE_PAREN,	0 },
	{ ")",		OP_RPAREN,	STAGE_PAREN,	0 }
};

typedef struct __token_class
{
	TOKEN_TYPE type;
	union
	{
		OP_TYPE operator;
		union
		{
			char *sval;
			long ival;
		} operand;
	} val;
} TOKEN;


static char opchars[] = "+-*/%#&|^=!<>~?:,()";

static void get_next_token (char **line, TOKEN *token)
{
	char *ptr = *line;
	char *next = ptr;

	while (ptr && *ptr && isspace(*ptr))
		next++, ptr++;

	/*
	 * Is it an operand?
	 */
	if (!strchr(opchars, *ptr))
	{
		if (*ptr == '[')
		{
			next = MatchingBracket(ptr + 1, LEFT_BRACKET, RIGHT_BRACKET);
			if (next)
				*next++ = 0;

			token->type = TOKEN_TEXT;
			token->val.operand.sval = ptr + 1;
		}
		else if (my_isdigit(ptr))
		{
			token->type = TOKEN_NUMBER;
			token->val.operand.ival = strtol(ptr, &next, 0);
		}
		else
		{
			next = ptr + 1;
			while (*next && !isspace(*next) && !strchr(opchars, *next))
				next++;

			token->val.operand.sval = ptr;
			token->type = TOKEN_VAR;

			if (*next == '(')
			{
				char *ptr2;

				*next = ' ';
				ptr2 = MatchingBracket(next + 1, LEFT_PAREN, RIGHT_PAREN);
				if (ptr2)
				{
					*ptr2++ = 0;
					next = ptr2;
					token->type = TOKEN_FUNCTION_CALL;
				}
				else
					next++;		/* XXXX */
			}

			/* Delimit if we can get away with it. */
			if (isspace(*next))
				*next++ = 0;
		}
	}
	else
	{
		token->type = TOKEN_OPERATOR;

		/* 
		 * There are some things we always do, for the sake of
		 * our own sanity.
		 *
		 * Because users can enter something like:
		 *	foo+=4
		 * (eg, no spaces), we cannot just start throwing around
		 * nulls nilly willy, as it will corrupt the next token.
		 * This really is only a problem for operators, since its
		 * not possible to have two consecutive operands without
		 * some kind of operator (or space) inbetween.
		 *
		 * So what we do is we do NOT null out the following char
		 * when we extract an operand, but instead we null out the
		 * first char of an operator once we have identified it.
		 * That way when we actually do go and parse the operand,
		 * it will be properly null terminated, and we can still
		 * have the correct operand.
		 */
		switch (ptr[0])
		{
			case ('+') :
			{
				next++;
				switch (ptr[1])
				{
					case ('+'):
						token->val.operator = OP_INCREMENT;
						next++;
						break;
					case ('='):
						token->val.operator = OP_PLUSEQ;
						next++;
						break;
					default:
						token->val.operator = OP_PLUS;
						break;
				}
				break;
			}
			case ('-') :
			{
				next++;
				switch (ptr[1])
				{
					case ('-'):
						token->val.operator = OP_DECREMENT;
						next++;
						break;
					case ('='):
						token->val.operator = OP_MINUSEQ;
						next++;
						break;
					default:
						token->val.operator = OP_MINUS;
						break;
				}
				break;
			}
			case ('*') :
			{
				next++;
				switch (ptr[1])
				{
					case ('='):
						token->val.operator = OP_MULTEQ;
						next++;
						break;
					default:
						token->val.operator = OP_MULT;
						break;
				}
				break;
			}
			case ('/') :
			{
				next++;
				switch (ptr[1])
				{
					case ('='):
						token->val.operator = OP_DIVEQ;
						next++;
						break;
					default:
						token->val.operator = OP_DIV;
						break;
				}
				break;
			}
			case ('%') :
			{
				next++;
				switch (ptr[1])
				{
					case ('='):
						token->val.operator = OP_MODEQ;
						next++;
						break;
					default:
						token->val.operator = OP_MOD;
						break;
				}
				break;
			}
			case ('#') :
			{
				next++;
				switch (ptr[1])
				{
					case ('='):
						token->val.operator = OP_STRCAT;
						next++;
						break;
					default:
						token->val.operator = OP_STRAPP;
						break;
				}
				break;
			}
			case ('&') :
			{
				next++;
				switch (ptr[1])
				{
					case ('&'):
						token->val.operator = OP_LOGAND;
						next++;
						break;
					case ('='):
						token->val.operator = OP_BITANDEQ;
						next++;
						break;
					default:
						token->val.operator = OP_BITAND;
						break;
				}
				break;
			}
			case ('|') :
			{
				next++;
				switch (ptr[1])
				{
					case ('|'):
						token->val.operator = OP_LOGOR;
						next++;
						break;
					case ('='):
						token->val.operator = OP_BITOREQ;
						next++;
						break;
					default:
						token->val.operator = OP_BITOR;
						break;
				}
				break;
			}
			case ('^') :
			{
				next++;
				switch (ptr[1])
				{
					case ('^'):
						token->val.operator = OP_LOGXOR;
						next++;
						break;
					case ('='):
						token->val.operator = OP_BITXOREQ;
						next++;
						break;
					default:
						token->val.operator = OP_BITXOR;
						break;
				}
				break;
			}
			case ('=') :
			{
				next++;
				switch (ptr[1])
				{
					case ('='):
						token->val.operator = OP_EQ;
						next++;
						break;
					default:
						token->val.operator = OP_ASSN;
						break;
				}
				break;
			}
			case ('!'):
			{
				next++;
				switch (ptr[1])
				{
					case ('='):
						token->val.operator = OP_NEQ;
						next++;
						break;
					default:
						token->val.operator = OP_NOT;
						break;
				}
				break;
			}
			case ('<'):
			{
				next++;
				switch (ptr[1])
				{
					case ('='):
						token->val.operator = OP_LTEQ;
						next++;
						break;
					default:
						token->val.operator = OP_LT;
						break;
				}
				break;
			}
			case ('>'):
			{
				next++;
				switch (ptr[1])
				{
					case ('='):
						token->val.operator = OP_GTEQ;
						next++;
						break;
					default:
						token->val.operator = OP_GT;
						break;
				}
				break;
			}
			case ('~'):
			{
				next++;
				token->val.operator = OP_COMP;
				break;
			}
			case ('?'):
			{
				next++;
				token->val.operator = OP_TERT1;
				break;
			}
			case (':'):
			{
				next++;
				token->val.operator = OP_TERT2;
				break;
			}
			case (','):
			{
				next++;
				token->val.operator = OP_COMMA;
				break;
			}
			case ('('):
			{
				next++;
				token->val.operator = OP_LPAREN;
				break;
			}
			case (')'):
			{
				next++;
				token->val.operator = OP_RPAREN;
				break;
			}
		}
		*ptr = 0;		/* delimit previous token */
	}
	*line = next;
}


/* Evaluates the sub expression from token 'low' to token 'high'. */
static char *evaluate_sub_expreesion (TOKEN *token, int low, int high)
{
	int c, d;
	int count;
	int stage;

	for (stage = STAGE_MULT; stage < STAGE_NONE; stage--)
	{
		for (c = low; c <= high; c++)
		{
			if (operator && operators[token[c].operator].stage == stage)
			{
				if (operator == binary)
				{
					get_left_op()
					get_right_op()
					perform_operation();
					put_result_in_left_op();
					kill_operator();
					kill_right_op();
				}
				else if (operator == unary)
				{
					if (assoc == ASSOC_LEFT)
					{
						get_left_op();
						perform_operation();
						put_result_in_left_op();
						kill_operator();
					}
					else if (assoc == ASSOC_RIGHT)
					{
						get_right_op();
						perform_operation();
						put_result_in_operator();
						kill_right_op();
					}
					else if (assoc == ASSOC_BOTH)
					{
						if (left_op() == operand && stage == STAGE_POSTFIX)
							get_left_op();
						else if (right_op() == operand && stage == STAGE_PREFIX)
							get_right_op();
						else
							ignore();

						perform_operation();
						put_result_in_operator();
						kill_some_op();
					}
				}
				else if (operator == tertiary)
				{
					find_corresponding_tert();
					get_left_op();
					if (left_op == true)
						get_left_op_of_tert();
					else
						get_right_op_of_tert();
				}
			}
		}
	}
}






#if 0
			if (token[c].type == TOKEN_OPERATOR
			  && token[c].val.operator == OPER_LPAREN)
			{
				for (d = c + 1; count && d < high; d++)
				{
					if (token[d].type == TOKEN_OPERATOR && token[d].val.operator == OPER_LPAREN)
						count++;
					else if (token[d].type == TOKEN_OPERATOR && token[d].val.operator == OPER_RPAREN)
						count--;
				}
				evaluate_sub_expression(token, c + 1, d - 1);
				token[c].type = TOKEN_NONE;
				token[d].type = TOKEN_NONE;
			}
#endif


static char *evaluate_expression (char *str)
{
	TOKEN token[1024];
	int i = 0;
	int c;

	while (str && *str)
		get_next_token(&str, &token[i]), i++;

	evaluate_sub_expression(token, 0, i - 1);
	return first_op();
}

int main (void)
{
	char buffer[256];

	fgets(buffer, 255, stdin);
	buffer[strlen(buffer)-1] = 0;

	while (*buffer)
	{
		evaluate_expression(buffer);
		fgets(buffer, 255, stdin);
		buffer[strlen(buffer)-1] = 0;
	}

	exit(0);
}

/* MatchingBracket returns the next unescaped bracket of the given type */
char     *MatchingBracket(char *string, char left, char right)
{
        int     bracket_count = 1;

        while (*string && bracket_count)
        {
                if (*string == left)
                        bracket_count++;
                else if (*string == right)
                {
                        if (!--bracket_count)
                                return string;
                }
                else if (*string == '\\' && string[1])
                        string++;
                string++;
        }
        return (char *) 0;
}

==> epic4pre0.041.tar.gz: ircii-EPIC4pre0.041/source/expr2.c <==
/*
 * A new math parser, such as it is.
 * This is still a work in progress.  It is provided
 * for your amusement and comments if you want.
 *
 * Written by Jeremy Nelson
 * Copyright 1997 EPIC Software Labs
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
char     *MatchingBracket(char *string, char left, char right);
#define LEFT_BRACKET '['
#define RIGHT_BRACKET ']'
#define LEFT_PAREN '('
#define RIGHT_PAREN ')'

#define my_isdigit(x) \
(*x >= '0' && *x <= '9') || \
((*x == '-'  || *x == '+') && (x[1] >= '0' && x[1] <= '9'))

typedef enum _token_type
{
	TOKEN_UNKNOWN,
	TOKEN_TEXT,
	TOKEN_VAR,
	TOKEN_NUMBER,
	TOKEN_OPERATOR,
	TOKEN_FUNCTION_CALL
} TOKEN_TYPE;

typedef enum _operator_type
{
	OP_UNKNOWN,
	OP_PLUS,	OP_MINUS,	OP_INCREMENT, 	OP_DECREMENT,
	OP_PLUSEQ,	OP_MINUSEQ,

	OP_MULT,	OP_DIV,		OP_MOD,		
	OP_MULTEQ,	OP_DIVEQ,	OP_MODEQ,

	OP_STRAPP,	OP_STRCAT,

	OP_BITAND,	OP_LOGAND,	OP_BITANDEQ,
	OP_BITOR,	OP_LOGOR,	OP_BITOREQ,
	OP_BITXOR,	OP_LOGXOR,	OP_BITXOREQ,

	OP_EQ,		OP_NEQ,		OP_ASSN,
	OP_LT,		OP_LTEQ,	OP_GT,		OP_GTEQ,
	OP_COMP,	OP_NOT,

	OP_TERT1,	OP_TERT2,	OP_COMMA,
	OP_LPAREN,	OP_RPAREN
} OP_TYPE;

#define	STAGE_NONE	0
#define STAGE_PAREN	1
#define	STAGE_COMMA	2
#define	STAGE_ASSN	3
#define	STAGE_TERT	4
#define	STAGE_CONJ	5
#define	STAGE_BITW	6
#define	STAGE_COMP	7
#define	STAGE_ADD	8
#define	STAGE_MULT	9

#define UNARY 1
#define ASSOC_BOTH 2
#define ASSOC_LEFT 4

struct __opvals
{
	char 		*op_str;
	OP_TYPE  	op_type;
	int		stage_type;
	int		flags;
} operators[] = 
{
	{ "",		OP_UNKNOWN,	STAGE_NONE,	0 },
	{ "+",		OP_PLUS,	STAGE_ADD,	0 },
	{ "-",		OP_MINUS,	STAGE_ADD,	0 },
	{ "++",		OP_INCREMENT,	STAGE_NONE,	UNARY | ASSOC_BOTH },
	{ "--",		OP_INCREMENT,	STAGE_NONE, 	UNARY | ASSOC_BOTH },
	{ "+=",		OP_PLUSEQ,	STAGE_ASSN, 	ASSOC_LEFT },
	{ "-=",		OP_MINUSEQ,	STAGE_ASSN, 	ASSOC_LEFT },
	{ "*",		OP_MULT,	STAGE_MULT, 	0 },
	{ "/",		OP_DIV,		STAGE_MULT, 	0 },
	{ "%",		OP_MOD,		STAGE_MULT, 	0 },
	{ "*=",		OP_MULTEQ,	STAGE_ASSN, 	ASSOC_LEFT },
	{ "/=",		OP_DIVEQ,	STAGE_ASSN, 	ASSOC_LEFT },
	{ "%=",		OP_MODEQ,	STAGE_ASSN, 	ASSOC_LEFT },
	{ "##",		OP_STRAPP,	STAGE_ADD, 	0 },
	{ "#=",		OP_STRCAT,	STAGE_ASSN, 	ASSOC_LEFT },
	{ "&",		OP_BITAND,	STAGE_BITW, 	0 },
	{ "&&",		OP_LOGAND,	STAGE_CONJ, 	0 },
	{ "&=",		OP_BITANDEQ,	STAGE_ASSN, 	ASSOC_LEFT },
	{ "|",		OP_BITOR,	STAGE_BITW, 	0 },
	{ "||",		OP_LOGOR,	STAGE_CONJ, 	0 },
	{ "|=",		OP_BITOREQ,	STAGE_ASSN, 	ASSOC_LEFT },
	{ "^",		OP_BITXOR,	STAGE_BITW, 	0 },
	{ "^^",		OP_LOGXOR,	STAGE_CONJ, 	0 },
	{ "^=",		OP_BITXOREQ,	STAGE_ASSN, 	ASSOC_LEFT },
	{ "==",		OP_EQ,		STAGE_COMP, 	0 },
	{ "!=",		OP_NEQ,		STAGE_COMP, 	0 },
	{ "=",		OP_ASSN,	STAGE_ASSN, 	ASSOC_LEFT },
	{ "<",		OP_LT,		STAGE_COMP, 	0 },
	{ ">",		OP_GT,		STAGE_COMP, 	0 },
	{ "<=",		OP_LTEQ,	STAGE_COMP, 	0 },
	{ ">=",		OP_GTEQ,	STAGE_COMP, 	0 },
	{ "~",		OP_COMP,	STAGE_BITW, 	UNARY },
	{ "!",		OP_COMP,	STAGE_CONJ,	UNARY },
	{ "?",		OP_TERT1,	STAGE_TERT,	0 },
	{ ":",		OP_TERT2,	STAGE_TERT,	0 },
	{ ",",		OP_COMMA,	STAGE_COMMA,	0 },
	{ "(",		OP_LPAREN,	STAGE_PAREN,	0 },
	{ ")",		OP_RPAREN,	STAGE_PAREN,	0 }
};

typedef struct __token_class
{
	TOKEN_TYPE type;
	union
	{
		OP_TYPE operator;
		union
		{
			char *sval;
			long ival;
		} operand;
	} val;
} TOKEN;


static char opchars[] = "+-*/%#&|^=!<>~?:,()";

static void get_next_token (char **line, TOKEN *token)
{
	char *ptr = *line;
	char *next = ptr;

	while (ptr && *ptr && isspace(*ptr))
		next++, ptr++;

	/*
	 * Is it an operand?
	 */
	if (!strchr(opchars, *ptr))
	{
		if (*ptr == '[')
		{
			next = MatchingBracket(ptr + 1, LEFT_BRACKET, RIGHT_BRACKET);
			if (next)
				*next++ = 0;

			token->type = TOKEN_TEXT;
			token->val.operand.sval = ptr + 1;
		}
		else if (my_isdigit(ptr))
		{
			token->type = TOKEN_NUMBER;
			token->val.operand.ival = strtol(ptr, &next, 0);
		}
		else
		{
			next = ptr + 1;
			while (*next && !isspace(*next) && !strchr(opchars, *next))
				next++;

			token->val.operand.sval = ptr;
			token->type = TOKEN_VAR;

			if (*next == '(')
			{
				char *ptr2;

				*next = ' ';
				ptr2 = MatchingBracket(next + 1, LEFT_PAREN, RIGHT_PAREN);
				if (ptr2)
				{
					*ptr2++ = 0;
					next = ptr2;
					token->type = TOKEN_FUNCTION_CALL;
				}
				else
					next++;		/* XXXX */
			}

			/* Delimit if we can get away with it. */
			if (isspace(*next))
				*next++ = 0;
		}
	}
	else
	{
		token->type = TOKEN_OPERATOR;

		/* 
		 * There are some things we always do, for the sake of
		 * our own sanity.
		 *
		 * Because users can enter something like:
		 *	foo+=4
		 * (eg, no spaces), we cannot just start throwing around
		 * nulls nilly willy, as it will corrupt the next token.
		 * This really is only a problem for operators, since its
		 * not possible to have two consecutive operands without
		 * some kind of operator (or space) inbetween.
		 *
		 * So what we do is we do NOT null out the following char
		 * when we extract an operand, but instead we null out the
		 * first char of an operator once we have identified it.
		 * That way when we actually do go and parse the operand,
		 * it will be properly null terminated, and we can still
		 * have the correct operand.
		 */
		switch (ptr[0])
		{
			case ('+') :
			{
				next++;
				switch (ptr[1])
				{
					case ('+'):
						token->val.operator = OP_INCREMENT;
						next++;
						break;
					case ('='):
						token->val.operator = OP_PLUSEQ;
						next++;
						break;
					default:
						token->val.operator = OP_PLUS;
						break;
				}
				break;
			}
			case ('-') :
			{
				next++;
				switch (ptr[1])
				{
					case ('-'):
						token->val.operator = OP_DECREMENT;
						next++;
						break;
					case ('='):
						token->val.operator = OP_MINUSEQ;
						next++;
						break;
					default:
						token->val.operator = OP_MINUS;
						break;
				}
				break;
			}
			case ('*') :
			{
				next++;
				switch (ptr[1])
				{
					case ('='):
						token->val.operator = OP_MULTEQ;
						next++;
						break;
					default:
						token->val.operator = OP_MULT;
						break;
				}
				break;
			}
			case ('/') :
			{
				next++;
				switch (ptr[1])
				{
					case ('='):
						token->val.operator = OP_DIVEQ;
						next++;
						break;
					default:
						token->val.operator = OP_DIV;
						break;
				}
				break;
			}
			case ('%') :
			{
				next++;
				switch (ptr[1])
				{
					case ('='):
						token->val.operator = OP_MODEQ;
						next++;
						break;
					default:
						token->val.operator = OP_MOD;
						break;
				}
				break;
			}
			case ('#') :
			{
				next++;
				switch (ptr[1])
				{
					case ('='):
						token->val.operator = OP_STRCAT;
						next++;
						break;
					default:
						token->val.operator = OP_STRAPP;
						break;
				}
				break;
			}
			case ('&') :
			{
				next++;
				switch (ptr[1])
				{
					case ('&'):
						token->val.operator = OP_LOGAND;
						next++;
						break;
					case ('='):
						token->val.operator = OP_BITANDEQ;
						next++;
						break;
					default:
						token->val.operator = OP_BITAND;
						break;
				}
				break;
			}
			case ('|') :
			{
				next++;
				switch (ptr[1])
				{
					case ('|'):
						token->val.operator = OP_LOGOR;
						next++;
						break;
					case ('='):
						token->val.operator = OP_BITOREQ;
						next++;
						break;
					default:
						token->val.operator = OP_BITOR;
						break;
				}
				break;
			}
			case ('^') :
			{
				next++;
				switch (ptr[1])
				{
					case ('^'):
						token->val.operator = OP_LOGXOR;
						next++;
						break;
					case ('='):
						token->val.operator = OP_BITXOREQ;
						next++;
						break;
					default:
						token->val.operator = OP_BITXOR;
						break;
				}
				break;
			}
			case ('=') :
			{
				next++;
				switch (ptr[1])
				{
					case ('='):
						token->val.operator = OP_EQ;
						next++;
						break;
					default:
						token->val.operator = OP_ASSN;
						break;
				}
				break;
			}
			case ('!'):
			{
				next++;
				switch (ptr[1])
				{
					case ('='):
						token->val.operator = OP_NEQ;
						next++;
						break;
					default:
						token->val.operator = OP_NOT;
						break;
				}
				break;
			}
			case ('<'):
			{
				next++;
				switch (ptr[1])
				{
					case ('='):
						token->val.operator = OP_LTEQ;
						next++;
						break;
					default:
						token->val.operator = OP_LT;
						break;
				}
				break;
			}
			case ('>'):
			{
				next++;
				switch (ptr[1])
				{
					case ('='):
						token->val.operator = OP_GTEQ;
						next++;
						break;
					default:
						token->val.operator = OP_GT;
						break;
				}
				break;
			}
			case ('~'):
			{
				next++;
				token->val.operator = OP_COMP;
				break;
			}
			case ('?'):
			{
				next++;
				token->val.operator = OP_TERT1;
				break;
			}
			case (':'):
			{
				next++;
				token->val.operator = OP_TERT2;
				break;
			}
			case (','):
			{
				next++;
				token->val.operator = OP_COMMA;
				break;
			}
			case ('('):
			{
				next++;
				token->val.operator = OP_LPAREN;
				break;
			}
			case (')'):
			{
				next++;
				token->val.operator = OP_RPAREN;
				break;
			}
		}
		*ptr = 0;		/* delimit previous token */
	}
	*line = next;
}


/* Evaluates the sub expression from token 'low' to token 'high'. */
static char *evaluate_sub_expression (TOKEN *token, int low, int high)
{
	int c, d;
	int count;
	int stage;

	for (stage = STAGE_MULT; stage < STAGE_NONE; stage--)
	{
		for (c = low; c <= high; c++)
		{
			if (operator && operators[token[c].operator].stage == stage)
			{
				if (operator == binary)
				{
					get_left_op()
					get_right_op()
					perform_operation();
					put_result_in_left_op();
					kill_operator();
					kill_right_op();
				}
				else if (operator == unary)
				{
					if (assoc == ASSOC_LEFT)
					{
						get_left_op();
						perform_operation();
						put_result_in_left_op();
						kill_operator();
					}
					else if (assoc == ASSOC_RIGHT)
					{
						get_right_op();
						perform_operation();
						put_result_in_operator();
						kill_right_op();
					}
					else if (assoc == ASSOC_BOTH)
					{
						if (left_op() == operand && stage == STAGE_POSTFIX)
							get_left_op();
						else if (right_op() == operand && stage == STAGE_PREFIX)
							get_right_op();
						else
							ignore();

						perform_operation();
						put_result_in_operator();
						kill_some_op();
					}
				}
				else if (operator == tertiary)
				{
					find_corresponding_tert();
					get_left_op();
					if (left_op == true)
						get_left_op_of_tert();
					else
						get_right_op_of_tert();
				}
			}
		}
	}
}






#if 0
			if (token[c].type == TOKEN_OPERATOR
			  && token[c].val.operator == OPER_LPAREN)
			{
				for (d = c + 1; count && d < high; d++)
				{
					if (token[d].type == TOKEN_OPERATOR && token[d].val.operator == OPER_LPAREN)
						count++;
					else if (token[d].type == TOKEN_OPERATOR && token[d].val.operator == OPER_RPAREN)
						count--;
				}
				evaluate_sub_expression(token, c + 1, d - 1);
				token[c].type = TOKEN_NONE;
				token[d].type = TOKEN_NONE;
			}
#endif


static char *evaluate_expression (char *str)
{
	TOKEN token[1024];
	int i = 0;
	int c;

	while (str && *str)
		get_next_token(&str, &token[i]), i++;

	evaluate_sub_expression(token, 0, i - 1);
	return first_op();
}

int main (void)
{
	char buffer[256];

	fgets(buffer, 255, stdin);
	buffer[strlen(buffer)-1] = 0;

	while (*buffer)
	{
		evaluate_expression(buffer);
		fgets(buffer, 255, stdin);
		buffer[strlen(buffer)-1] = 0;
	}

	exit(0);
}

/* MatchingBracket returns the next unescaped bracket of the given type */
char     *MatchingBracket(char *string, char left, char right)
{
        int     bracket_count = 1;

        while (*string && bracket_count)
        {
                if (*string == left)
                        bracket_count++;
                else if (*string == right)
                {
                        if (!--bracket_count)
                                return string;
                }
                else if (*string == '\\' && string[1])
                        string++;
                string++;
        }
        return (char *) 0;
}
